/*
 * Copyright 2006, U.S. Robotics Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND U.S. Robotics GRANTS NO WARRANTIES OF
 * ANY KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. U.S.
 * Robotics SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */
/*
 * Firewall services
 *
 * Copyright 2006, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <dirent.h>
/* not used yet
#include <errno.h>
#include <linux/sockios.h>
#include <sys/ioctl.h>
#include <unistd.h>
*/
#include <bcmnvram.h>
#include <shutils.h>
#include <rc.h>
#include <netconf.h>
#include <nvparse.h>

/* Get list of IP Networks */
/*not used yet
char *
get_lanlist(void)
{
	int i;
	static char buf[255];
	char ifname[32],*val=NULL;
	
	memset(buf,0,sizeof(buf));
	
	for (i = 0; i < 16; i++){
		snprintf(ifname,sizeof(ifname),"lan%d_ifname",i);
		val = nvram_get(ifname);
		if (val)
			snprintf(buf,sizeof(buf),"%s%s%s",buf,(*buf) ? " ": "",val);
	}
	
	if (*buf) 
		return buf;
	else
		return NULL;
	
	
}	
*/
	
/* Add filter to specified table */
static void
add_filter(netconf_filter_t *filter, int dir)
{
	filter->dir = dir;
	netconf_add_filter(filter);
}


/* Add port forward and a matching ACCEPT rule to the FORWARD table */
static void
add_forward(netconf_nat_t *nat, int dir, int target)
{
	netconf_filter_t filter;

	/* Set up LAN side match */
	memset(&filter, 0, sizeof(filter));
	filter.match.ipproto = nat->match.ipproto;
	filter.match.src.ports[1] = nat->match.src.ports[1];
	filter.match.dst.ipaddr.s_addr = nat->ipaddr.s_addr;
	filter.match.dst.netmask.s_addr = htonl(0xffffffff);
	filter.match.dst.ports[0] = nat->ports[0];
	filter.match.dst.ports[1] = nat->ports[1];
	strncpy(filter.match.in.name, nat->match.in.name, IFNAMSIZ);

	/* Accept connection */
	filter.target = target;
	filter.dir = dir;

	/* Do it */
	netconf_add_nat(nat);
	netconf_add_filter(&filter);
}

/* KEO - need to delete pptp firewall rules */
/* Add port forward and a matching ACCEPT rule to the FORWARD table */
static void
del_forward(netconf_nat_t *nat, int dir, int target)
{
	netconf_filter_t filter;

	/* Set up LAN side match */
	memset(&filter, 0, sizeof(filter));
	filter.match.ipproto = nat->match.ipproto;
	filter.match.src.ports[1] = nat->match.src.ports[1];
	filter.match.dst.ipaddr.s_addr = nat->ipaddr.s_addr;
	filter.match.dst.netmask.s_addr = htonl(0xffffffff);
	filter.match.dst.ports[0] = nat->ports[0];
	filter.match.dst.ports[1] = nat->ports[1];
	strncpy(filter.match.in.name, nat->match.in.name, IFNAMSIZ);

	/* Accept connection */
	filter.target = target;
	filter.dir = dir;

	/* Do it */
	netconf_del_nat(nat);
	netconf_del_filter(&filter);
}


int
start_firewall(void)
{
	DIR *dir;
	struct dirent *file;
	FILE *fp;
	char name[NAME_MAX];
	netconf_filter_t filter;
	netconf_app_t app;
	int i;
	char var[256], *next;
	int log_level, log_drop, log_accept;
	
	/*not used yet
	struct ipaddress_pairs {
				struct in_addr address;
				struct in_addr mask;
				} *ip_addrs = NULL;
	char *interfaces;
	*/
	/* Reset firewall */
	netconf_reset_fw();

	/* Block obviously spoofed IP addresses */
	if (!(dir = opendir("/proc/sys/net/ipv4/conf")))
		perror("/proc/sys/net/ipv4/conf");
	while (dir && (file = readdir(dir))) {
		if (strncmp(file->d_name, ".", NAME_MAX) != 0 &&
		    strncmp(file->d_name, "..", NAME_MAX) != 0) {
			snprintf(name, sizeof(name), "/proc/sys/net/ipv4/conf/%s/rp_filter", file->d_name);
			if (!(fp = fopen(name, "r+"))) {
				perror(name);
				break;
			}
			fputc('1', fp);
			fclose(fp);
		}
	}
	if (dir)
		closedir(dir);

	/* Optionally log connections */
	log_level = atoi(nvram_safe_get("log_level"));
	log_drop = (log_level & 1) ? NETCONF_LOG_DROP : NETCONF_DROP;
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on")) {
		log_drop = NETCONF_LOG_DROP;
		log_accept = NETCONF_LOG_ACCEPT;
	}
	/*
	 * Inbound drops
	 */

	/* Drop invalid packets */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_INVALID;
	filter.target = NETCONF_DROP;
	add_filter(&filter, NETCONF_IN);
	add_filter(&filter, NETCONF_FORWARD);

	/*
	 * Forward drops
	 */

	/* Client filters */
	for (i = 0; i < MAX_NVPARSE; i++) {
		netconf_filter_t start, end;

		if (get_filter_client(i, &start, &end) && !(start.match.flags & NETCONF_DISABLED)) {
			while (ntohl(start.match.src.ipaddr.s_addr) <= ntohl(end.match.src.ipaddr.s_addr)) {
				/* Override target */
				start.target = log_drop;
				add_filter(&start, NETCONF_FORWARD);
				start.match.src.ipaddr.s_addr = htonl(ntohl(start.match.src.ipaddr.s_addr) + 1);
			}
		}
	}

	/* Filter by MAC address */
	if (!nvram_match("filter_macmode", "disabled")) {
		memset(&filter, 0, sizeof(filter));
		strncpy(filter.match.in.name, nvram_safe_get("lan_ifname"), IFNAMSIZ);
		if (nvram_match("filter_macmode", "allow")) {
			/* Allow new connections from the LAN side only from the specified addresses */
			filter.target = log_accept;
			filter.match.state = NETCONF_NEW;
		} else {
			/* Drop connections from the specified addresses */
			filter.target = log_drop;
		}
		foreach(var, nvram_safe_get("filter_maclist"), next) {
			if (ether_atoe(var, filter.match.mac.octet))
				add_filter(&filter, NETCONF_FORWARD);
		}
	}

	/*
	 * Inbound accepts
	 */
	/* Allow new connections from the loopback interface */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_NEW;
	/* not used yet
	filter.match.src.ipaddr.s_addr = htonl(INADDR_LOOPBACK);
	filter.match.src.netmask.s_addr = htonl(INADDR_BROADCAST);
	*/
	strncpy(filter.match.in.name, "lo",IFNAMSIZ);
	filter.target = log_accept;
	add_filter(&filter, NETCONF_IN);
	add_filter(&filter, NETCONF_FORWARD);

	/* Allow new connections from the LAN side */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_NEW;
	strncpy(filter.match.in.name, nvram_safe_get("lan_ifname"),IFNAMSIZ);
	/* not used yet
	inet_aton(nvram_safe_get("lan_ipaddr"),(struct in_addr *)&filter.match.src.ipaddr);
	inet_aton(nvram_safe_get("lan_netmask"),(struct in_addr *)&filter.match.src.netmask);
	filter.match.src.ipaddr.s_addr &= filter.match.src.netmask.s_addr;	
	*/
	filter.target = log_accept;
	add_filter(&filter, NETCONF_IN);
	if (!nvram_match("filter_macmode", "allow"))
		add_filter(&filter, NETCONF_FORWARD);


	/* Allow DHCP broadcast requests */
	/* not used yet
	if (nvram_match("lan_proto","dhcp")){
		memset(&filter, 0, sizeof(filter));
		filter.match.state = NETCONF_NEW;
		filter.match.ipproto = IPPROTO_UDP;
		strncpy(filter.match.in.name, nvram_safe_get("lan_ifname"),IFNAMSIZ);
		filter.match.src.netmask.s_addr = htonl(INADDR_BROADCAST);
		filter.match.src.ipaddr.s_addr = htonl(INADDR_ANY);
		filter.match.dst.netmask.s_addr = htonl(INADDR_BROADCAST);
		filter.match.dst.ipaddr.s_addr = htonl(INADDR_BROADCAST);
		filter.match.dst.ports[0]=htons(67);
		filter.match.dst.ports[1]=htons(67);
		filter.match.src.ports[0]=htons(68);
		filter.match.src.ports[1]=htons(68);
		filter.target = log_accept;
		add_filter(&filter, NETCONF_IN);
	};

	//Allow new connections from unbridged interfaces		
	
	interfaces = get_lanlist();
	if (interfaces){
	
		char *address,*mask;
		int num_ipaddrs;
		char word[64], *next,*ifname;
		char wan_if[32],*wan_ifname;
		char lan_ip[32],lan_ipmask[32],lan_proto[32];
		int index ;
		struct ipaddress_pairs *cur_ipaddr,lan_ipaddr;
			
		// Get the list IP addresses of the bridged/internal and the guest interfaces 
		// This builds the lists use to program the restriction filters, ie the
		// ones that isolate the guest SSID lans from each other and the internal network.
		
		
		num_ipaddrs=0;
		
		if (nvram_get("lan_ipaddr")) num_ipaddrs++;
			
		// Count the number of guest interfaces	
		foreach(word,interfaces,next) 
			num_ipaddrs++;
		
		ip_addrs = malloc(sizeof(struct ipaddress_pairs) * (num_ipaddrs));
		
		if (!ip_addrs) {
			dprintf("start_firewall()Error allocating memory\n");
			goto start_firewall_cleanup0;
		}
		
		memset(ip_addrs,0,sizeof(struct ipaddress_pairs) * (num_ipaddrs));
		
		// Get bridge ip and netmask info from NVRAM 
		
		address=nvram_get("lan_ipaddr");
		mask=nvram_get("lan_netmask");
		cur_ipaddr=ip_addrs;
		
		if (address && mask){
			if (!inet_aton(address,&cur_ipaddr->address)){
				cprintf("start_firewall():Invalid lan_ipaddr:%s\n",address);
				goto start_firewall_cleanup1;
			}			
			if (!inet_aton(mask,&cur_ipaddr->mask)){
				cprintf("start_firewall():Invalid lan_netmask:%s\n",mask);
				goto start_firewall_cleanup1;
			}
			cur_ipaddr++;			
			
		}else if (address || mask) { // If either mask or ip values are present but not both we have a problem
			dprintf("start_firewall():Both bridge ipaddress and ipmask must be defined.\n");
			goto start_firewall_cleanup1;
		}
		
		// Get guest LAN ip address and netmask info from NVRAM
		
		interfaces = get_lanlist();
		foreach(word,interfaces,next){
			index = get_ipconfig_index(word);
			
			// Drop the interfaces that have no valid IP config
			if (index < 0) 
				continue;
			
			if (index > MAX_NVPARSE){
				cprintf("start_firewall():Interface index(%d) exceeds MAX_NVPARSE(%d)\n",
										index,MAX_NVPARSE);
				goto start_firewall_cleanup1;
			}
			
			snprintf(lan_ip,sizeof(lan_ip),"lan%d_ipaddr",index);
			snprintf(lan_ipmask,sizeof(lan_ip),"lan%d_netmask",index);
			address=nvram_get(lan_ip);
			mask=nvram_get(lan_ipmask);
		
			if (address && mask){

				if (!inet_aton(address,&cur_ipaddr->address)){
					cprintf("start_firewall():Invalid %s:%s\n",lan_ip,address);
					goto start_firewall_cleanup1;
				}
							
				if (!inet_aton(mask,&cur_ipaddr->mask)){
					cprintf("start_firewall():Invalid %s:%s\n",lan_ipmask,mask);
					goto start_firewall_cleanup1;
				cur_ipaddr++;
				}
			}else{
				cprintf("start_firewall():%s(val=%s) and/or %s(val=%s) not defined.\n",
										lan_ip,lan_ipmask,address,mask);
				goto start_firewall_cleanup1;
			}
	
		}
		
		// Start building the filter rules
		interfaces = get_lanlist();
		foreach(word,interfaces,next){

			ifname = word;
			index = get_ipconfig_index(ifname);
			
			if (index < 0 ) continue;
			if ((index > MAX_NVPARSE) || (index < 0)){
				cprintf("start_firewall():Index (%d) exceeds MAX_NVPARSE(%d)\n",
										index,MAX_NVPARSE);
				goto start_firewall_cleanup1;
			}
					
			// The filter rules drops packets to all the internal interfaces. 
			// * The interfaces were enumerated earlier. 
			// * We exclude the interface's own ip address and that of its 
			// * wan corresponding gateway.
			// * The wan interface is pointed by the NVRAM variable lanXX_wan_if
			
				
			snprintf(lan_ip,sizeof(wan_if),"lan%d_ipaddr",index);
			snprintf(lan_ipmask,sizeof(wan_if),"lan%d_netmask",index);
			snprintf(wan_if,sizeof(wan_if),"lan%d_wan_if",index);
			snprintf(lan_proto,sizeof(lan_proto),"lan%d_proto",index);
			
			// Get the LAN IP Address/Netmask from NVRAM			
			address=nvram_get(lan_ip);
			mask=nvram_get(lan_ipmask);
			wan_ifname=nvram_get(wan_if);
		
			if (address && mask){
				if (!inet_aton(address,&lan_ipaddr.address)){
					cprintf("start_firewall():Invalid %s:%s\n",lan_ip,address);
					goto start_firewall_cleanup1;
				}			
				if (!inet_aton(mask,&lan_ipaddr.mask)){
					cprintf("start_firewall():Invalid %s:%s\n",lan_ipmask,mask);
					goto start_firewall_cleanup1;
				}
			}else{
				dprintf("start_firewall():%s(val=%s) and/or %s(val=%s) not defined.\n",
										lan_ip,lan_ipmask,address,mask);
				goto start_firewall_cleanup1;
			}

			// Apply restriction filters to internal IP addresses
			for (index=0;index < num_ipaddrs;  index++){
				// Skip making rule if the dest address is ours 
				// Assume that the ranges do not overlap so we do not
				// check the mask value
				
				if ( lan_ipaddr.address.s_addr == 
						ip_addrs[index].address.s_addr) continue;
				
				//Create the restriction filter
				memset(&filter, 0, sizeof(filter));			
				filter.match.state = NETCONF_NEW;
				filter.match.src.ipaddr.s_addr = lan_ipaddr.address.s_addr;
				filter.match.src.ipaddr.s_addr &= lan_ipaddr.mask.s_addr;
				filter.match.src.netmask.s_addr = lan_ipaddr.mask.s_addr;
				filter.match.dst.ipaddr.s_addr = ip_addrs[index].address.s_addr;
				filter.match.dst.ipaddr.s_addr &= ip_addrs[index].mask.s_addr;
				filter.match.dst.netmask.s_addr = ip_addrs[index].mask.s_addr;		
				strncpy(filter.match.in.name, ifname,IFNAMSIZ);
				filter.target = log_drop;
				add_filter(&filter, NETCONF_IN);
				add_filter(&filter, NETCONF_FORWARD);		
			}
							
			// If DHCP is configured allow broadcasts to DHCP server port (67) from client DHCP port (68)	
			if (nvram_match(lan_proto,"dhcp")){
				memset(&filter, 0, sizeof(filter));
				filter.match.state = NETCONF_NEW;
				filter.match.ipproto = IPPROTO_UDP;
				strncpy(filter.match.in.name, ifname,IFNAMSIZ);
				filter.match.src.netmask.s_addr = htonl(INADDR_BROADCAST);
				filter.match.src.ipaddr.s_addr = htonl(INADDR_ANY);
				filter.match.dst.netmask.s_addr = htonl(INADDR_BROADCAST);
				filter.match.dst.ipaddr.s_addr = htonl(INADDR_BROADCAST);
				filter.match.dst.ports[0]=htons(67);
				filter.match.dst.ports[1]=htons(67);
				filter.match.src.ports[0]=htons(68);
				filter.match.src.ports[1]=htons(68);
				filter.target = log_accept;
				add_filter(&filter, NETCONF_IN);
			}
					
			memset(&filter,0 ,sizeof(filter));
			filter.match.state = NETCONF_NEW;
			filter.match.src.ipaddr.s_addr = lan_ipaddr.address.s_addr;
			filter.match.src.ipaddr.s_addr &= lan_ipaddr.mask.s_addr;
			filter.match.src.netmask.s_addr = lan_ipaddr.mask.s_addr;
			strncpy(filter.match.in.name, ifname,IFNAMSIZ);
			filter.target = log_accept;
			add_filter(&filter, NETCONF_IN);
			
			if (!nvram_match("filter_macmode", "allow")){
				//strncpy(filter.match.out.name, nvram_safe_get(wan_if),IFNAMSIZ);
				add_filter(&filter, NETCONF_FORWARD);	
			}					
						
		}
		free(ip_addrs);
		ip_addrs = NULL;
		
	}
	*/
	/* Allow established and related outbound connections back in */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_ESTABLISHED | NETCONF_RELATED;
	filter.target = log_accept;
	add_filter(&filter, NETCONF_IN);
	add_filter(&filter, NETCONF_FORWARD);

	/*
	 * NAT
	 */

	/* Enable IP masquerading */
	if ((fp = fopen("/proc/sys/net/ipv4/ip_forward", "r+"))) {
		fputc('1', fp);
		fclose(fp);
	} else
		perror("/proc/sys/net/ipv4/ip_forward");

	/* Application specific port forwards */
	for (i = 0; i < MAX_NVPARSE; i++) {
		memset(&app, 0, sizeof(app));
		if (get_autofw_port(i, &app) && !(app.match.flags & NETCONF_DISABLED))
			netconf_add_fw((netconf_fw_t *) &app);
	}

	/*
	* Inbound defaults
	*/
	
	/* Log refused connections */
	memset(&filter, 0, sizeof(filter));
	filter.target = log_drop;
	add_filter(&filter, NETCONF_IN);
	add_filter(&filter, NETCONF_FORWARD);
	
	dprintf("done\n");
	return 0;
/* not used yet
start_firewall_cleanup1:
	if (ip_addrs){
	 free(ip_addrs);
	 ip_addrs = NULL;
	}
	
start_firewall_cleanup0:
	return 1;
*/
}


int
start_firewall2(char *wan_ifname)
{
	netconf_nat_t nat;
	netconf_filter_t filter;
	int log_level, log_accept;
	char tmp[100], prefix[] = "wanXXXXXXXXXX_";
	int i;
		
	/* Optionally log connections */
	log_level = atoi(nvram_safe_get("log_level"));
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on"))
		log_accept = NETCONF_LOG_ACCEPT;
	/* Allow new connections from the WAN side */
	if (nvram_match("fw_disable", "1")) {
		memset(&filter, 0, sizeof(filter));
		filter.match.state = NETCONF_NEW;
		strncpy(filter.match.in.name, wan_ifname, IFNAMSIZ);
		filter.target = log_accept;
		add_filter(&filter, NETCONF_IN);
		add_filter(&filter, NETCONF_FORWARD);
	}

	/* Enable IP masquerading */
	memset(&nat, 0, sizeof(nat));
	inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
	inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
	nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	strncpy(nat.match.out.name, wan_ifname, IFNAMSIZ);
	nat.target = NETCONF_MASQ;
	netconf_add_nat(&nat);
	
	/* Enable IP masquerading for unbridged interfaces*/
	/* not used yet
	{
	char *interfaces, *ifname;
	char word[64], *next;
	int index =1 ;
	
	interfaces = get_lanlist();
	if (interfaces)
		foreach(word,interfaces,next)
		{
		 	char interface[32], mask[32], wan[32];
			
			ifname = word;
			index = get_ipconfig_index(ifname);
			
			if (index < 0 )
				continue;
						
			snprintf(interface,sizeof(interface),"lan%d_ipaddr",index);
			snprintf(wan,sizeof(wan),"wan%d_ipaddr",index);
			snprintf(mask,sizeof(mask),"lan%d_netmask",index);
			
			memset(&nat, 0, sizeof(nat));
			inet_aton(nvram_safe_get(interface), &nat.match.src.ipaddr);
			inet_aton(nvram_safe_get(mask), &nat.match.src.netmask);
			nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
			strncpy(nat.match.out.name, wan_ifname,IFNAMSIZ);
			nat.target = NETCONF_MASQ;
			netconf_add_nat(&nat);	
		}
	}	
	*/
	/* Enable remote management */
	if (nvram_invmatch("http_wanport", "") && nvram_invmatch("wan_pptp_activated","1")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[0] = htons(atoi(nvram_safe_get("http_wanport")));
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("http_wanport")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[0] = htons(atoi(nvram_safe_get("http_lanport")));
		nat.ports[1] = htons(atoi(nvram_safe_get("http_lanport")));

		add_forward(&nat, NETCONF_IN, log_accept);
	}
	
	/* Enable remote printing */
	if (nvram_match("wan_ipp", "1") && nvram_invmatch("wan_pptp_activated","1")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));

		add_forward(&nat, NETCONF_IN, log_accept);
	}

	/* Persistent (static) port forwards */
	for (i = 0; i < MAX_NVPARSE; i++) {
		memset(&nat, 0, sizeof(nat));
		if (get_forward_port(i, &nat) && !(nat.match.flags & NETCONF_DISABLED)) {
			/* Set interface name (match packets entering WAN interface) */
			strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);
			add_forward(&nat, NETCONF_FORWARD, log_accept);
		}
	}

	/* Forward all WAN ports to DMZ IP address */
	if (nvram_invmatch("dmz_ipaddr", "")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(0xffff);
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("dmz_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(0xffff);

		nat.match.ipproto = IPPROTO_TCP;
		add_forward(&nat, NETCONF_FORWARD, log_accept);
		nat.match.ipproto = IPPROTO_UDP;
		add_forward(&nat, NETCONF_FORWARD, log_accept);
	}

	/* Clamp TCP MSS to PMTU of WAN interface */
	snprintf(prefix,sizeof(prefix), "wan%d_", wan_ifunit(wan_ifname));
	if (nvram_match(strcat_r(prefix, "proto", tmp), "pppoe"))
		netconf_clamp_mss_to_pmtu();

	dprintf("done\n");
	return 0;
}

static void delete_persistent_port_forwards(char *wan_ifname)
{
	netconf_nat_t nat;
	int log_level, log_accept;
	int i;
	
	// Optionally log connections 
	log_level = atoi(nvram_safe_get("log_level"));
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on"))
		log_accept = NETCONF_LOG_ACCEPT;
		
	for (i = 0; i < MAX_NVPARSE; i++) {
	    memset(&nat, 0, sizeof(nat));
	    if (get_forward_port(i, &nat) && !(nat.match.flags & NETCONF_DISABLED)) {
	    	/* Set interface name (match packets entering WAN interface) */
		strncpy(nat.match.in.name, nvram_get("wan_ifname"), IFNAMSIZ);
		del_forward(&nat, NETCONF_FORWARD, log_accept);
	    }
	}
}


static void add_persistent_port_forwards(char *wan_ifname)
{
	netconf_nat_t nat;
	int log_level, log_accept;
	int i;
	
	// Optionally log connections 
	log_level = atoi(nvram_safe_get("log_level"));
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on"))
		log_accept = NETCONF_LOG_ACCEPT;
		
	for (i = 0; i < MAX_NVPARSE; i++) {
	    memset(&nat, 0, sizeof(nat));
	    if (get_forward_port(i, &nat) && !(nat.match.flags & NETCONF_DISABLED)) {
		/* Set interface name (match packets entering WAN interface) */
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);
		add_forward(&nat, NETCONF_FORWARD, log_accept);
	    }
	}
}

int
start_firewall3(char *wan_ifname, char *local, char *remote, char *mask)
{

	netconf_nat_t nat;
	int log_level, log_accept;
	netconf_filter_t filter;
		
	// Optionally log connections 
	log_level = atoi(nvram_safe_get("log_level"));
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on"))
		log_accept = NETCONF_LOG_ACCEPT;
		
	/* Allow new connections from the WAN side */
	if (nvram_match("fw_disable", "1")) {
		memset(&filter, 0, sizeof(filter));
		filter.match.state = NETCONF_NEW;
		strcpy(filter.match.in.name, wan_ifname);
		filter.target = log_accept;
		add_filter(&filter, NETCONF_IN);
		add_filter(&filter, NETCONF_FORWARD);
	}
	
	// Enable IP masquerading 
	memset(&nat, 0, sizeof(nat));
	inet_aton(remote, &nat.match.src.ipaddr);

	inet_aton(mask, &nat.match.src.netmask);
	nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	strcpy(nat.match.out.name, wan_ifname);
	nat.target = NETCONF_MASQ;
	netconf_add_nat(&nat);

	if (nvram_invmatch("wan_proto","pppoe"))
	{
    	    // Disable the old NAT 
	    memset(&nat, 0, sizeof(nat));
	    inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
	    inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
	    nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	    strcpy(nat.match.out.name, nvram_get("wan_ifname"));
	    nat.target = NETCONF_MASQ;
	    netconf_del_nat(&nat);

	    /* kill default route */    
	    route_del(nvram_get("wan_ifname"), 0
			, "0.0.0.0", 
			nvram_safe_get("wan0_gateway"), 
			"0.0.0.0");

	    cprintf("route del default gw %s dev %s\n", 
		    nvram_safe_get("wan0_gateway"), nvram_get("wan_ifname"));

    	    // All pass bt the tunnel 
	    memset(&nat, 0, sizeof(nat));
	    inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
	    inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
	    nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	    strcpy(nat.match.out.name, wan_ifname);
	    nat.target = NETCONF_MASQ;
	    netconf_add_nat(&nat);

	    /* add new default route through pptp tunnel*/    
	    cprintf("%s : add default route to %s\n",wan_ifname,local); 
	    route_add(wan_ifname, 0
			, "0.0.0.0", 
			local, 
			"0.0.0.0");

	    /* add new route for the tunnel*/    
	    cprintf("%s : add static route %s to dev %s\n",wan_ifname,
		    local,nvram_get("wan_ifname")); 

	    route_add(nvram_get("wan_ifname"), 0
			, local, 
			"0.0.0.0",
			"255.255.255.255");

	    /* Delete Persistent (static) port forwards to WAN interface */
	    delete_persistent_port_forwards(nvram_get("wan_ifname"));
	    /* ADD Persistent (static) port forwards to PPTP interface */
	    add_persistent_port_forwards(wan_ifname);
	    
	    /* Enable remote management */
	    if (nvram_invmatch("http_wanport", "")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[0] = htons(atoi(nvram_safe_get("http_wanport")));
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("http_wanport")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[0] = htons(atoi(nvram_safe_get("http_lanport")));
		nat.ports[1] = htons(atoi(nvram_safe_get("http_lanport")));

		add_forward(&nat, NETCONF_IN, log_accept);
	    }
	    
	    /* Enable remote printing */
	    if (nvram_match("wan_ipp", "1")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));

		add_forward(&nat, NETCONF_IN, log_accept);
	    }
	    
	    /* Forward all PPTP ports to DMZ IP address */
	    if (nvram_invmatch("dmz_ipaddr", "")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(0xffff);
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("dmz_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(0xffff);

		nat.match.ipproto = IPPROTO_TCP;
		add_forward(&nat, NETCONF_FORWARD, log_accept);
		nat.match.ipproto = IPPROTO_UDP;
		add_forward(&nat, NETCONF_FORWARD, log_accept);
	    }
	}
	
	return 0;
}

int
stop_firewall3(char *wan_ifname, char *local, char *remote, char *mask)
{

	netconf_nat_t nat;
	int log_level, log_accept;
	
	// Optionally log connections 
	log_level = atoi(nvram_safe_get("log_level"));
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;
	if (nvram_match("sysdbg","on"))
		log_accept = NETCONF_LOG_ACCEPT;

	// Enable IP masquerading 
	memset(&nat, 0, sizeof(nat));
	inet_aton(remote, &nat.match.src.ipaddr);
	inet_aton(mask, &nat.match.src.netmask);
	nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	strcpy(nat.match.out.name, wan_ifname);
	nat.target = NETCONF_MASQ;
	netconf_del_nat(&nat);

	if (nvram_invmatch("wan_proto","pppoe"))
	{
	    memset(&nat, 0, sizeof(nat));
	    inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
	    inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
	    nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	    strcpy(nat.match.out.name, wan_ifname);
	    nat.target = NETCONF_MASQ;
	    netconf_del_nat(&nat);

	    /* del new route for the tunnel*/    
	    cprintf("%s : del static route %s to dev %s\n",wan_ifname,
		    local,nvram_get("wan_ifname")); 
	    route_del(nvram_get("wan_ifname"), 0
			, local, 
			"0.0.0.0",
			"255.255.255.255");

	    /* del default route to pptp tunnel*/    
	    cprintf("%s : del default route to %s\n",wan_ifname,local); 
	    route_del(wan_ifname, 0
			, "0.0.0.0", 
			local, 
			"0.0.0.0");

	     	
    	    // Enable IP masquerading 
	    memset(&nat, 0, sizeof(nat));
	    inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
	    inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
	    nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
	    strcpy(nat.match.out.name, nvram_get("wan_ifname"));
	    nat.target = NETCONF_MASQ;
	    netconf_add_nat(&nat);

	    /* add default route back */    
	    cprintf("%s : restore default route %s to dev %s\n",wan_ifname,
		    nvram_get("wan0_gateway"),nvram_get("wan_ifname")); 
	    route_add(nvram_get("wan_ifname"), 0
			, "0.0.0.0", 
			nvram_safe_get("wan0_gateway"), 
			"0.0.0.0");

	    /* Delete Persistent (static) port forwards to PPTP interface */
	    delete_persistent_port_forwards(wan_ifname);
	    /* Add Persistent (static) port forwards to WAN interface */
	    add_persistent_port_forwards(nvram_get("wan_ifname"));
	  
	    /* Disable remote management */
	    if (nvram_invmatch("http_wanport", "")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[0] = htons(atoi(nvram_safe_get("http_wanport")));
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("http_wanport")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[0] = htons(atoi(nvram_safe_get("http_lanport")));
		nat.ports[1] = htons(atoi(nvram_safe_get("http_lanport")));

		del_forward(&nat, NETCONF_IN, log_accept);
	    }
	    
	    /* Disable remote printing */
	    if (nvram_match("ipp_port", "1")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = IPPROTO_TCP;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("lan_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(atoi(nvram_safe_get("ipp_port")));

		del_forward(&nat, NETCONF_IN, log_accept);
	    }
	      
	    /* delete forward rules from PPTP ports to DMZ IP address */
	    if (nvram_invmatch("dmz_ipaddr", "")) {
		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[1] = htons(0xffff);
		strncpy(nat.match.in.name, wan_ifname, IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		(void) inet_aton(nvram_safe_get("dmz_ipaddr"), &nat.ipaddr);
		nat.ports[1] = htons(0xffff);

		nat.match.ipproto = IPPROTO_TCP;
		del_forward(&nat, NETCONF_FORWARD, log_accept);
		nat.match.ipproto = IPPROTO_UDP;
		del_forward(&nat, NETCONF_FORWARD, log_accept);
	    }
	}

	return 0;
}
